@@ -1,6 +1,7 @@
 package aws
 
 import (
+	"context"
 	"fmt"
 	"strings"
 	"time"
@@ -8,6 +9,7 @@ import (
 	"github.com/aws/aws-sdk-go/aws"
 	"github.com/aws/aws-sdk-go/aws/arn"
 	"github.com/aws/aws-sdk-go/aws/awserr"
+	"github.com/aws/aws-sdk-go/aws/endpoints"
 	"github.com/aws/aws-sdk-go/aws/request"
 	"github.com/aws/aws-sdk-go/aws/session"
 	"github.com/aws/aws-sdk-go/service/ec2"
@@ -72,15 +74,18 @@ func New(logger logrus.FieldLogger, metadata *types.ClusterMetadata) (providers.
 	for _, filter := range metadata.ClusterPlatformMetadata.AWS.Identifier {
 		filters = append(filters, filter)
 	}
-
-	session, err := awssession.GetSession()
+	region := metadata.ClusterPlatformMetadata.AWS.Region
+	session, err := awssession.GetSessionWithOptions(
+		awssession.WithRegion(region),
+		awssession.WithServiceEndpoints(region, metadata.ClusterPlatformMetadata.AWS.ServiceEndpoints),
+	)
 	if err != nil {
 		return nil, err
 	}
 
 	return &ClusterUninstaller{
 		Filters:   filters,
-		Region:    metadata.ClusterPlatformMetadata.AWS.Region,
+		Region:    region,
 		Logger:    logger,
 		ClusterID: metadata.InfraID,
 		Session:   session,
@@ -101,16 +106,13 @@ func (o *ClusterUninstaller) Run() error {
 		return err
 	}
 
-	awsConfig := &aws.Config{Region: aws.String(o.Region)}
 	awsSession := o.Session
 	if awsSession == nil {
 		// Relying on appropriate AWS ENV vars (eg AWS_PROFILE, AWS_ACCESS_KEY_ID, etc)
-		awsSession, err = session.NewSession(awsConfig)
+		awsSession, err = session.NewSession(aws.NewConfig().WithRegion(o.Region))
 		if err != nil {
 			return err
 		}
-	} else {
-		awsSession = awsSession.Copy(awsConfig)
 	}
 	awsSession.Handlers.Build.PushBackNamed(request.NamedHandler{
 		Name: "openshiftInstaller.OpenshiftInstallerUserAgentHandler",
@@ -123,12 +125,21 @@ func (o *ClusterUninstaller) Run() error {
 	tagClientNames := map[*resourcegroupstaggingapi.ResourceGroupsTaggingAPI]string{
 		tagClients[0]: o.Region,
 	}
-	if o.Region != "us-east-1" {
-		tagClient := resourcegroupstaggingapi.New(
-			awsSession, aws.NewConfig().WithRegion("us-east-1"),
-		)
-		tagClients = append(tagClients, tagClient)
-		tagClientNames[tagClient] = "us-east-1"
+
+	switch o.Region {
+	case endpoints.CnNorth1RegionID, endpoints.CnNorthwest1RegionID:
+		if o.Region != endpoints.CnNorthwest1RegionID {
+			tagClient := resourcegroupstaggingapi.New(awsSession, aws.NewConfig().WithRegion(endpoints.CnNorthwest1RegionID))
+			tagClients = append(tagClients, tagClient)
+			tagClientNames[tagClient] = endpoints.CnNorthwest1RegionID
+		}
+
+	default:
+		if o.Region != endpoints.UsEast1RegionID {
+			tagClient := resourcegroupstaggingapi.New(awsSession, aws.NewConfig().WithRegion(endpoints.UsEast1RegionID))
+			tagClients = append(tagClients, tagClient)
+			tagClientNames[tagClient] = endpoints.UsEast1RegionID
+		}
 	}
 
 	iamClient := iam.New(awsSession)
@@ -148,12 +159,14 @@ func (o *ClusterUninstaller) Run() error {
 		return err
 	}
 
+	tracker := new(errorTracker)
+	tagClientStack := append([]*resourcegroupstaggingapi.ResourceGroupsTaggingAPI(nil), tagClients...)
 	err = wait.PollImmediateInfinite(
 		time.Second*10,
 		func() (done bool, err error) {
 			var loopError error
 			nextTagClients := tagClients[:0]
-			for _, tagClient := range tagClients {
+			for _, tagClient := range tagClientStack {
 				matched := false
 				for _, filter := range o.Filters {
 					o.Logger.Debugf("search for and delete matching resources by tag in %s matching %#+v", tagClientNames[tagClient], filter)
@@ -178,14 +191,9 @@ func (o *ClusterUninstaller) Run() error {
 										continue
 									}
 
-									if parsed.Service == "ec2" && strings.HasPrefix(parsed.Resource, "network-interface/") {
-										arnLogger.Debug("skip direct tag deletion, deleting the VPC should catch this")
-										continue
-									}
-
 									err = deleteARN(awsSession, parsed, filter, arnLogger)
 									if err != nil {
-										arnLogger.Debug(err)
+										tracker.suppressWarning(arnString, err, arnLogger)
 										err = errors.Wrapf(err, "deleting %s", arnString)
 										continue
 									}
@@ -210,7 +218,7 @@ func (o *ClusterUninstaller) Run() error {
 					o.Logger.Debugf("no deletions from %s, removing client", tagClientNames[tagClient])
 				}
 			}
-			tagClients = nextTagClients
+			tagClientStack = nextTagClients
 
 			o.Logger.Debug("search for IAM roles")
 			arns, err := iamRoleSearch.arns()
@@ -242,7 +250,7 @@ func (o *ClusterUninstaller) Run() error {
 
 					err = deleteARN(awsSession, parsed, nil, arnLogger)
 					if err != nil {
-						arnLogger.Debug(err)
+						tracker.suppressWarning(arnString, err, arnLogger)
 						err = errors.Wrapf(err, "deleting %s", arnString)
 						loopError = err
 						continue
@@ -251,7 +259,7 @@ func (o *ClusterUninstaller) Run() error {
 				}
 			}
 
-			return len(tagClients) == 0 && loopError == nil, nil
+			return len(tagClientStack) == 0 && loopError == nil, nil
 		},
 	)
 	if err != nil {
@@ -263,6 +271,12 @@ func (o *ClusterUninstaller) Run() error {
 		o.Logger.Debug(err)
 		return err
 	}
+
+	err = removeSharedTags(context.TODO(), tagClients, tagClientNames, o.Filters, o.Logger)
+	if err != nil {
+		return err
+	}
+
 	return nil
 }
 
@@ -555,6 +569,10 @@ func deleteEC2(session *session.Session, arn arn.ARN, filter Filter, logger logr
 		return deleteEC2Volume(client, id, logger)
 	case "vpc":
 		return deleteEC2VPC(client, elb.New(session), elbv2.New(session), id, logger)
+	case "vpc-endpoint":
+		return deleteEC2VPCEndpoint(client, id, logger)
+	case "vpc-peering-connection":
+		return deleteEC2VPCPeeringConnection(client, id, logger)
 	default:
 		return errors.Errorf("unrecognized EC2 resource type %s", resourceType)
 	}
@@ -679,7 +697,7 @@ func terminateEC2InstanceByInstance(ec2Client *ec2.EC2, iamClient *iam.IAM, inst
 		return err
 	}
 
-	logger.Info("Terminating")
+	logger.Debug("Terminating")
 	return nil
 }
 
@@ -690,6 +708,11 @@ func terminateEC2InstancesByTags(ec2Client *ec2.EC2, iamClient *iam.IAM, filters
 		return nil, errors.New("EC2 client does not have region configured")
 	}
 
+	partition, ok := endpoints.PartitionForRegion(endpoints.DefaultPartitions(), *ec2Client.Config.Region)
+	if !ok {
+		return nil, errors.Errorf("no partition found for region %q", *ec2Client.Config.Region)
+	}
+
 	terminated := map[string]struct{}{}
 	err := wait.PollImmediateInfinite(
 		time.Second*10,
@@ -720,9 +743,9 @@ func terminateEC2InstancesByTags(ec2Client *ec2.EC2, iamClient *iam.IAM, filters
 
 								instanceLogger := logger.WithField("instance", *instance.InstanceId)
 								if *instance.State.Name == "terminated" {
-									arn := fmt.Sprintf("arn:aws:ec2:%s:%s:instance/%s", *ec2Client.Config.Region, *reservation.OwnerId, *instance.InstanceId)
+									arn := fmt.Sprintf("arn:%s:ec2:%s:%s:instance/%s", partition.ID(), *ec2Client.Config.Region, *reservation.OwnerId, *instance.InstanceId)
 									if _, ok := terminated[arn]; !ok {
-										instanceLogger.Debug("Terminated")
+										instanceLogger.Info("Terminated")
 										terminated[arn] = exists
 									}
 									continue
@@ -1147,6 +1170,7 @@ func deleteEC2SubnetsByVPC(client *ec2.EC2, vpc string, failFast bool, logger lo
 		return err
 	}
 
+	var lastError error
 	for _, subnet := range results.Subnets {
 		err := deleteEC2Subnet(client, *subnet.SubnetId, logger.WithField("subnet", *subnet.SubnetId))
 		if err != nil {
@@ -1154,10 +1178,14 @@ func deleteEC2SubnetsByVPC(client *ec2.EC2, vpc string, failFast bool, logger lo
 			if failFast {
 				return err
 			}
+			if lastError != nil {
+				logger.Debug(lastError)
+			}
+			lastError = err
 		}
 	}
 
-	return nil
+	return lastError
 }
 
 func deleteEC2Volume(client *ec2.EC2, id string, logger logrus.FieldLogger) error {
@@ -1255,6 +1283,20 @@ func deleteEC2VPCEndpointsByVPC(client *ec2.EC2, vpc string, failFast bool, logg
 	return nil
 }
 
+func deleteEC2VPCPeeringConnection(client *ec2.EC2, id string, logger logrus.FieldLogger) error {
+	_, err := client.DeleteVpcPeeringConnection(&ec2.DeleteVpcPeeringConnectionInput{
+		VpcPeeringConnectionId: &id,
+	})
+	if err != nil {
+		if err.(awserr.Error).Code() == "InvalidVpcPeeringConnectionID.NotFound" {
+			return nil
+		}
+		return errors.Wrapf(err, "cannot delete VPC Peering Connection %s", id)
+	}
+	logger.Info("Deleted")
+	return nil
+}
+
 func deleteElasticLoadBalancing(session *session.Session, arn arn.ARN, logger logrus.FieldLogger) error {
 	resourceType, id, err := splitSlash("resource", arn.Resource)
 	if err != nil {
@@ -1683,6 +1725,11 @@ func deleteRoute53(session *session.Session, arn arn.ARN, logger logrus.FieldLog
 
 	sharedZoneID, err := getSharedHostedZone(client, id, logger)
 	if err != nil {
+		// In some cases AWS may return the zone in the list of tagged resources despite the fact
+		// it no longer exists.
+		if err.(awserr.Error).Code() == route53.ErrCodeNoSuchHostedZone {
+			return nil
+		}
 		return err
 	}
 
@@ -1794,6 +1841,45 @@ func deleteS3(session *session.Session, arn arn.ARN, logger logrus.FieldLogger)
 	}
 	logger.Debug("Emptied")
 
+	var lastError error
+	err = client.ListObjectVersionsPagesWithContext(aws.BackgroundContext(), &s3.ListObjectVersionsInput{
+		Bucket:  aws.String(arn.Resource),
+		MaxKeys: aws.Int64(1000),
+	}, func(page *s3.ListObjectVersionsOutput, lastPage bool) bool {
+		var deleteObjects []*s3.ObjectIdentifier
+		for _, deleteMarker := range page.DeleteMarkers {
+			deleteObjects = append(deleteObjects, &s3.ObjectIdentifier{
+				Key:       aws.String(*deleteMarker.Key),
+				VersionId: aws.String(*deleteMarker.VersionId),
+			})
+		}
+		for _, version := range page.Versions {
+			deleteObjects = append(deleteObjects, &s3.ObjectIdentifier{
+				Key:       aws.String(*version.Key),
+				VersionId: aws.String(*version.VersionId),
+			})
+		}
+		if len(deleteObjects) > 0 {
+			_, err := client.DeleteObjects(&s3.DeleteObjectsInput{
+				Bucket: aws.String(arn.Resource),
+				Delete: &s3.Delete{
+					Objects: deleteObjects,
+				},
+			})
+			if err != nil {
+				lastError = errors.Wrapf(err, "delete object failed %v", err)
+			}
+		}
+		return !lastPage
+	})
+	if lastError != nil {
+		return lastError
+	}
+	if err != nil && !isBucketNotFound(err) {
+		return err
+	}
+	logger.Debug("Versions Deleted")
+
 	_, err = client.DeleteBucket(&s3.DeleteBucketInput{
 		Bucket: aws.String(arn.Resource),
 	})
@@ -1826,3 +1912,85 @@ func isBucketNotFound(err interface{}) bool {
 	}
 	return false
 }
+
+func removeSharedTags(ctx context.Context, tagClients []*resourcegroupstaggingapi.ResourceGroupsTaggingAPI, tagClientNames map[*resourcegroupstaggingapi.ResourceGroupsTaggingAPI]string, filters []Filter, logger logrus.FieldLogger) error {
+	for _, filter := range filters {
+		for key, value := range filter {
+			if strings.HasPrefix(key, "kubernetes.io/cluster/") {
+				if value == "owned" {
+					if err := removeSharedTag(ctx, tagClients, tagClientNames, key, logger); err != nil {
+						return err
+					}
+				} else {
+					logger.Warnf("Ignoring non-owned cluster key %s: %s for shared-tag removal", key, value)
+				}
+			}
+		}
+	}
+
+	return nil
+}
+
+func removeSharedTag(ctx context.Context, tagClients []*resourcegroupstaggingapi.ResourceGroupsTaggingAPI, tagClientNames map[*resourcegroupstaggingapi.ResourceGroupsTaggingAPI]string, key string, logger logrus.FieldLogger) error {
+	request := &resourcegroupstaggingapi.UntagResourcesInput{
+		TagKeys: []*string{aws.String(key)},
+	}
+
+	removed := map[string]struct{}{}
+	tagClients = append([]*resourcegroupstaggingapi.ResourceGroupsTaggingAPI(nil), tagClients...)
+	for len(tagClients) > 0 {
+		nextTagClients := tagClients[:0]
+		for _, tagClient := range tagClients {
+			logger.Debugf("Search for and remove tags in %s matching %s: shared", tagClientNames[tagClient], key)
+			arns := []string{}
+			err := tagClient.GetResourcesPages(
+				&resourcegroupstaggingapi.GetResourcesInput{TagFilters: []*resourcegroupstaggingapi.TagFilter{{
+					Key:    aws.String(key),
+					Values: []*string{aws.String("shared")},
+				}}},
+				func(results *resourcegroupstaggingapi.GetResourcesOutput, lastPage bool) bool {
+					for _, resource := range results.ResourceTagMappingList {
+						arn := *resource.ResourceARN
+						if _, ok := removed[arn]; !ok {
+							arns = append(arns, arn)
+						}
+					}
+
+					return !lastPage
+				},
+			)
+			if err != nil {
+				err = errors.Wrap(err, "get tagged resources")
+				logger.Info(err)
+				nextTagClients = append(nextTagClients, tagClient)
+				continue
+			}
+			if len(arns) == 0 {
+				logger.Debugf("No matches in %s for %s: shared, removing client", tagClientNames[tagClient], key)
+				continue
+			}
+			nextTagClients = append(nextTagClients, tagClient)
+
+			for i := 0; i < len(arns); i += 20 {
+				request.ResourceARNList = make([]*string, 0, 20)
+				for j := 0; i+j < len(arns) && j < 20; j++ {
+					request.ResourceARNList = append(request.ResourceARNList, aws.String(arns[i+j]))
+				}
+				_, err = tagClient.UntagResourcesWithContext(ctx, request)
+				if err != nil {
+					err = errors.Wrap(err, "untag shared resources")
+					logger.Info(err)
+					continue
+				}
+				for j := 0; i+j < len(arns) && j < 20; j++ {
+					arn := arns[i+j]
+					logger.WithField("arn", arn).Infof("Removed tag %s: shared", key)
+					removed[arn] = exists
+				}
+			}
+		}
+		tagClients = nextTagClients
+	}
+
+	return nil
+}
